package PktDump::Analyser::DNS;

use strict;
use warnings;

use base qw( PktDump::Analyser );

use constant UDP_PORT => 53;

my @opcode_names = qw( QUERY IQUERY STATUS );
my @rcode_names  = qw( NOERROR FORMATERROR SERVFAIL NAMEERROR NOTIMPLEMENTED REFUSED );

my @class_names = qw( * IN CS CH HS ); $class_names[255] = "ANY";
my @type_names = qw(
   * A NS MD MF CNAME SOA MB
   MG MR NULL WKS PTR HINFO MINFO MX
   TXT RP AFSDB X25 ISDN RT NSAP NSAP_PTR
   SIG KEY PX GPOS AAAA LOC NXT EID
   NIMLOC SRV ATMA NAPTR KX CERT A6 DNAME
   SINK OPT APL );
$type_names[255] = "ANY";

sub analyse
{
   my $self = shift;
   my ( $pkt, $opts, $outer ) = @_;

   my ( $id, $flags, $qdcount, $ancount, $nscount, $arcount, $data ) = unpack( "n n n n n n a*", $pkt );

   my $qr     = $flags  & 1<<15;
   my $opcode = ($flags & 15<<11) >> 11;
   my $aa     = $flags  & 1<<10;
   my $tc     = $flags  & 1<<9;
   my $rd     = $flags  & 1<<8;
   my $ra     = $flags  & 1<<7;
   my $rcode  = ($flags & 15<<0) >> 0;

   if( $qr == 0 ) {
      # Query
      my $line = sprintf "DNS| %s->%s %s", $outer->{src}, $outer->{dst},
         $opcode_names[$opcode] || "opcode=$opcode";
      $line .= " id=$id" if $opts->{verbose};

      my @qd_rrs = map { $self->_extract_qd_rr( $data, $pkt, $opts ) } 1..$qdcount;

      if( $opts->{one_line} ) {
         return $line . ": " . join( " ", map { "[$_]" } @qd_rrs );
      }
      else {
         return $line . " query=$qdcount",;
            ( map { "DNS QD| $_" } @qd_rrs );
      }
   }
   else {
      # Response
      my $line = sprintf "DNS| %s->%s %s", $outer->{src}, $outer->{dst},
         $rcode_names[$rcode] || "rcode=$rcode";
      $line .= " id=$id" if $opts->{verbose};

      my @qd_rrs = map { $self->_extract_qd_rr( $data, $pkt, $opts ) } 1..$qdcount;
      my @an_rrs = map { $self->_extract_rr( $data, $pkt, $opts ) } 1..$ancount;
      my @ns_rrs = map { $self->_extract_rr( $data, $pkt, $opts ) } 1..$nscount;
      my @ar_rrs = map { $self->_extract_rr( $data, $pkt, $opts ) } 1..$arcount;

      if( $opts->{one_line} ) {
         return $line . ": " . join( " ", map { "[$_]" } @qd_rrs )
            . " => " . join( " ", map { "[$_]" } @an_rrs );
      }
      else {
         return $line . " answer=$ancount ns=$nscount additional=$arcount",
            ( map { "DNS QD| $_" } @qd_rrs ),
            ( map { "DNS AN| $_" } @an_rrs ),
            ( map { "DNS NS| $_" } @ns_rrs ),
            ( map { "DNS AR| $_" } @ar_rrs );
      }
   }
}

sub _extract_name
{
   my $self = shift;
   my ( undef, $pkt ) = @_;

   my $dclen = unpack "C", $_[0];
   if( ( $dclen & 0xc0 ) == 0xc0 ) {
      my $pos = unpack "n", $_[0]; substr( $_[0], 0, 2, "" );
      $pos &= 0x3fff;
      return $self->_extract_name( substr( $pkt, $pos ), $pkt );
   }
   else {
      substr( $_[0], 0, 1, "" );
      return "" unless $dclen;
      return substr( $_[0], 0, $dclen, "" ) . "." . $self->_extract_name( $_[0], $pkt );
   }
}

sub _extract_qd_rr
{
   my $self = shift;
   my ( undef, $pkt ) = @_;

   my $name = $self->_extract_name( $_[0], $pkt );
   my ( $qtype, $qclass ) = unpack( "n n", substr( $_[0], 0, 4, "" ) );

   return "$name $type_names[$qtype] $class_names[$qclass]";
}

sub _extract_rr
{
   my $self = shift;
   my ( undef, $pkt, $opts ) = @_;

   my $name = $self->_extract_name( $_[0], $pkt );
   my ( $type, $class, $ttl, $rdlen ) = unpack( "n n N n", substr( $_[0], 0, 10, "" ) );
   my $rrdata = substr( $_[0], 0, $rdlen, "" );

   my $rr = "$name $type_names[$type] $class_names[$class]";
   $rr .= " TTL $ttl" if $opts->{verbose};

   if( my $code = $self->can( "_extract_rr_$type_names[$type]" ) ) {
      $rr .= " " . $code->( $self, $rrdata, $pkt, $opts );
   }

   return $rr;
}

sub _extract_rr_dname
{
   my $self = shift;
   my ( undef, $pkt, $opts ) = @_;
   return $self->_extract_name( $_[0], $pkt );
}
*_extract_rr_CNAME = \&_extract_rr_dname;
*_extract_rr_NS = \&_extract_rr_dname;
*_extract_rr_PTR = \&_extract_rr_dname;

sub _extract_rr_A
{
   my $self = shift;
   my ( undef, $pkt, $opts ) = @_;
   return join ".", unpack( "C4", substr( $_[0], 0, 4, "" ) );
}

sub _extract_rr_AAAA
{
   my $self = shift;
   my ( undef, $pkt, $opts ) = @_;
   return join ":", map sprintf( "%x", $_ ), unpack( "n8", substr( $_[0], 0, 16, "" ) );
}

sub _extract_rr_MX
{
   my $self = shift;
   my ( undef, $pkt, $opts ) = @_;
   my ( $prio ) = unpack( "n", substr( $_[0], 0, 2, "" ) );
   my $name = $self->_extract_name( $_[0], $pkt );
   return "$prio $name";
}

sub _extract_rr_SRV
{
   my $self = shift;
   my ( undef, $pkt, $opts ) = @_;
   my ( $prio, $weight, $port ) = unpack( "n n n", substr( $_[0], 0, 6, "" ) );
   my $name = $self->_extract_name( $_[0], $pkt );
   return "$weight $prio $port $name";
}

1;
