package PktDump::Analyser::IPv6;

use strict;
use warnings;

use base qw( PktDump::Analyser );

use constant ETHERTYPE => 0x86dd;

use Socket 1.95 qw( AF_INET6 pack_sockaddr_in6 inet_ntop getnameinfo );

sub analyse
{
   my $self = shift;
   my ( $pkt, $opts ) = @_;

   my ( $vc ) = unpack( "C", $pkt );
   ( $vc & 0xf0 ) == 0x60 or die "Expected 6\n";

   my ( $vcl, $len, $nh, $hl, $src, $dst ) = unpack(
       "N     n     C    C    a16   a16", $pkt );
   my $data = substr $pkt, 40;

   my $class = ( $vcl >> 20 ) & 0xff;
   my $label = $vcl & 0xfffff;

   $_ = $self->format_ip6( $_, $opts ) for $src, $dst;

   my $line = sprintf( "IPv6| %s->%s: nh=%02x", $src, $dst, $nh );
   $line .= " hl=$hl" if $opts->{verbose};

   my $proto = $nh;

   my @lines;
   push @lines, $line unless $opts->{one_line};

   my $inner_analyser = eval { $self->for_ipproto( $proto ) };
   if( $@ and $@ =~ m/^Unable to find an Analyser for IPPROTO / ) {
      my @lines = $line;
      return $lines[0] if $opts->{one_line};

      push @lines, $self->format_hex( $data ) if $opts->{hexdump};

      return @lines;
   }
   elsif( $@ ) {
      die $@;
   }

   return @lines, $inner_analyser->analyse( $data, $opts, { src => $src, dst => $dst } );
}

my %namecache;
sub format_ip6
{
   my $self = shift;
   my ( $addr, $opts ) = @_;

   if( $opts->{resolve_hosts} ) {
      return $namecache{$addr} if exists $namecache{$addr};

      my ( $err, $name ) = getnameinfo( pack_sockaddr_in6( 0, $addr ), 0 );
      $name = inet_ntop AF_INET6, $addr if $err or $name =~ m/\.ip6\.arpa$/;
      return $namecache{$addr} = $name;
   }

   return inet_ntop AF_INET6, $addr;
}

*format_this_protaddr = \&format_ip6;

sub gen_opttype_addrmask { shift; PktDump::Analyser::IPv6::AddrMaskOptType->new( @_ ) }
{
   package PktDump::Analyser::IPv6::AddrMaskOptType;
   use base qw( PktDump::OptType::Directional PktDump::OptType );

   use Socket 1.95 qw( AF_INET6 inet_pton );

   use constant match => qr{[^ /]+(?i:/\d+)?};
   use constant desc  => q{$ADDR[/$SUBNET]};
   sub parse {
      my ( $host, $prefix ) = m{^([^ /]+)(?:/(\d+))?$};
      defined $prefix or $prefix = 128;

      $prefix <= 128 or die "Prefix length must be 0-128\n";

      my $hostaddr = inet_pton( AF_INET6, $host );
      defined $hostaddr or die "Unknown host '$host'\n";

      # Can only test in 32 bit chunks
      my @chunks;

      while( $prefix > 0 ) {
         ( my $bits, $hostaddr ) = unpack( "N a*", $hostaddr );

         my $chunklen = $prefix;
         $chunklen = 32 if $chunklen > 32;
         $prefix -= 32;

         push @chunks, [ $bits ] and next if $chunklen == 32;

         my $mask = 0xffffffff ^ ( ( 1 << ( 32 - $chunklen ) ) - 1 );
         $bits &= $mask;

         push @chunks, [ $bits, mask => $mask ];
      }

      return \@chunks;
   }
}

sub handle_filter_ipv6
   :Filter_description(Internet Protocol v6 (IPv6))
   :Filter_option(host => 'addrmask')
{
   my $self = shift;
   my ( $opts ) = @_;

   my $below = $self->build_layer_filter( proto => ETHERTYPE );

   my @parts = $below;

   if( my $host = $opts->{host} ) {
      my ( $dir, $hostspec ) = @$host;
      push @parts, $self->build_dir_filter( $dir,
         PktDump::Filter->all( map { $below->match_u32(  8 + 4*$_, @{ $hostspec->[$_] } ) } 0 .. $#$hostspec ),
         PktDump::Filter->all( map { $below->match_u32( 24 + 4*$_, @{ $hostspec->[$_] } ) } 0 .. $#$hostspec ),
      );
   }

   return PktDump::Filter->all( @parts );
}

1;
