package PktDump::Analyser::TCP;

use strict;
use warnings;

use base qw( PktDump::Analyser );

use constant IPPROTO => 6;

use PktDump::Filter;

use Socket 1.95 qw( INADDR_ANY pack_sockaddr_in getnameinfo NI_NUMERICHOST );

sub analyse
{
   my $self = shift;
   my ( $pkt, $opts, $outer ) = @_;

   my ( $offs ) = unpack( "x12 C", $pkt );
   $offs >>= 2; # >> 4 then * 4

   my ( $hdr, $data ) = unpack( "a$offs a*", $pkt );

   my ( $src, $dst, $seq, $ack, $flags, $win, $cksum, $urg, $tcpopts ) = unpack(
       "n     n     N     N    xC       n     n       n     a*", $hdr );

   $_ = $self->format_port( $_, $opts ) for $src, $dst;

   if( $opts->{one_line} ) {
      $src = "$outer->{src}:$src";
      $dst = "$outer->{dst}:$dst";
   }

   my $line = "TCP| ${src}->${dst}:";

   $line .= " seq=$seq" if $opts->{verbose};

   $line .= " URG($urg)" if $flags & 0x20;
   $line .= " ACK($ack)" if $flags & 0x10 and $opts->{verbose};
   $line .= " ACK"       if $flags & 0x10 and !$opts->{verbose};
   $line .= " PSH"       if $flags & 0x08;
   $line .= " RST"       if $flags & 0x04;
   $line .= " SYN"       if $flags & 0x02;
   $line .= " FIN"       if $flags & 0x01;

   $line .= sprintf " %d bytes", length $data if $flags & 0x08;

   my @lines = $line;
   push @lines, $self->format_hex( $data ) if $opts->{hexdump};

   return @lines;
}

my %namecache;
sub format_port
{
   my $self = shift;
   my ( $port, $opts ) = @_;

   if( $opts->{resolve_ports} ) {
      return $namecache{$port} if exists $namecache{$port};

      my ( $err, undef, $name ) = getnameinfo( pack_sockaddr_in( $port, INADDR_ANY ), NI_NUMERICHOST );
      return $namecache{$port} = $name;
   }

   return $port;
}

sub gen_opttype_dirports { shift; PktDump::Analyser::TCP::PortOptType->new( @_ ) }
{
   package PktDump::Analyser::TCP::PortOptType;
   use base qw( PktDump::OptType::Directional PktDump::OptType::MultiValue );

   sub parse_one
   {
      my $self = shift;
      my ( $port ) = @_;

      return $port if $port =~ m/^\d+$/;;

      my $portnum = getservbyname $port, 'tcp' or
         die "Cannot parse TCP port spec '$port' - $!";

      return $portnum;
   }
}

sub handle_filter_tcp
   :Filter_description(Transmission Control Protocol)
   :Filter_option(port => 'dirports')
{
   my $self = shift;
   my ( $opts ) = @_;

   my @parts = my $below = $self->build_layer_filter( ipproto => IPPROTO );

   if( my $port = $opts->{port} ) {
      my ( $dir, $ports ) = @$port;

      push @parts, $self->build_dir_filter( $dir,
         $below->match_u16( 0, $ports ), # src
         $below->match_u16( 2, $ports ), # dst
      );
   }

   return PktDump::Filter->all( @parts );
}

1;
