package PktDump::Analyser::UDP;

use strict;
use warnings;

use base qw( PktDump::Analyser );

use constant IPPROTO => 17;

use PktDump::Filter;

use Socket 1.95 qw( INADDR_ANY pack_sockaddr_in getnameinfo NI_NUMERICHOST NI_DGRAM );

sub analyse
{
   my $self = shift;
   my ( $pkt, $opts, $outer ) = @_;

   my ( $src, $dst, $len, $cksum, $data ) = unpack( "n n n n a*", $pkt );

   my $inner_analyser = eval { $self->for_udp_port( $dst ) } ||
                        eval { $self->for_udp_port( $src ) };
   my $err = $@;

   $_ = $self->format_port( $_, $opts ) for $src, $dst;

   if( $opts->{one_line} ) {
      $src = "$outer->{src}:$src";
      $dst = "$outer->{dst}:$dst";
   }

   my $line = sprintf "UDP| %s->%s %d bytes", $src, $dst, length $data;

   my @lines;
   push @lines, $line unless $opts->{one_line};

   if( $err and $err =~ m/^Unable to find an Analyser for UDP_PORT / ) {
      return $line if $opts->{one_line};
   
      push @lines, $self->format_hex( $data ) if $opts->{hexdump};

      return @lines;
   }
   elsif( $err ) {
      die $err;
   }

   return @lines, $inner_analyser->analyse( $data, $opts, { src => $src, dst => $dst } );
}

my %namecache;
sub format_port
{
   my $self = shift;
   my ( $port, $opts ) = @_;

   if( $opts->{resolve_ports} ) {
      return $namecache{$port} if exists $namecache{$port};

      my ( $err, undef, $name ) = getnameinfo( pack_sockaddr_in( $port, INADDR_ANY ), NI_NUMERICHOST|NI_DGRAM );
      return $namecache{$port} = $name;
   }

   return $port;
}

sub gen_opttype_dirports { shift; PktDump::Analyser::UDP::PortOptType->new( @_ ) }
{
   package PktDump::Analyser::UDP::PortOptType;
   use base qw( PktDump::OptType::Directional PktDump::OptType::MultiValue );

   sub parse_one
   {
      my $self = shift;
      my ( $port ) = @_;

      return $port if $port =~ m/^\d+$/;;

      my $portnum = getservbyname $port, 'udp' or
         die "Cannot parse UDP port spec '$port' - $!";

      return $portnum;
   }
}

sub handle_filter_udp
   :Filter_description(User Datagram Protocol)
   :Filter_option(port => 'dirports')
{
   my $self = shift;
   my ( $opts ) = @_;

   my @parts = my $below = $self->build_layer_filter( ipproto => IPPROTO );

   if( my $port = $opts->{port} ) {
      my ( $dir, $ports ) = @$port;

      push @parts, $self->build_dir_filter( $dir,
         $below->match_u16( 0, $ports ), # src
         $below->match_u16( 2, $ports ), # dst
      );
   }

   return PktDump::Filter->all( @parts );
}

1;
