package PktDump::Capfile::PcapNg;

# See also
#   http://www.winpcap.org/ntar/draft/PCAP-DumpFileFormat.html

use strict;
use warnings;

use base qw( PktDump::Capfile );
use constant FORMAT => "pcapng";

use Socket::Packet qw(
   PACKET_HOST PACKET_BROADCAST PACKET_MULTICAST PACKET_OTHERHOST PACKET_OUTGOING
);

use PktDump::Analyser;
use PktDump::Pack64 qw( $fmt_s64 pack_s64 unpack_s64 );

use constant {
   TYPE_SHB => 0x0A0D0D0A, # Section Header Block
   TYPE_IDB => 0x00000001, # Interface Description Block
   TYPE_EPB => 0x00000006, # Enhanced Packet Block
};

use constant PCAPNG_MAGIC => 0x1A2B3C4D;

use constant {
   IF_NAME    => 2, # Interface name
   IF_TSRESOL => 9, # Interface timestamp resolution
};

use constant {
   EPB_FLAGS => 2, # Packet flags
};

sub pack_block
{
   my $self = shift;
   my ( $type, $fmt, @args ) = @_;

   my $content = pack( $fmt, @args );

   my $len = (3 + length $content) & ~3; # round up to multiple of 4

   $len += 12; # account for type + len*2

   $self->pack( "i i a* x![i] i", $type, $len, $content, $len );
}

sub read_block
{
   my $self = shift;

   my ( $type, $len ) = $self->unpack( "i i", 8 )
      or return;
   $len -= 12; # account for type + len*2

   my ( $content ) = $self->unpack( "a*", $len );
   $self->unpack( "i", 4 );

   return $type, $content;
}

sub unpack_block
{
   my $self = shift;
   my ( $fmt ) = @_;

   my ( $type, $content ) = $self->read_block;

   return $type, unpack( $fmt, $content );
}

sub _pack_options
{
   my $self = shift;

   my $ret = "";
   while( @_ ) {
      my $type = shift;
      my $val = shift;

      $ret .= pack( "s s a* x![i]", $type, length $val, $val );
   }

   $ret .= pack( "s s", 0, 0 ); # opt_endofopt

   return $ret;
}

sub _unpack_options
{
   my $self = shift;

   ( local $_ ) = @_;

   my %opts;

   while( length ) {
      my ( $type, $len ) = unpack( "s s", substr( $_, 0, 4, "" ) );
      last if $type == 0;

      $opts{$type} = substr( $_, 0, $len, "" );
      substr( $_, 0, 4-($len%4), "" ) if $len%4;
   }

   return %opts;
}

sub write_header
{
   my $self = shift;

   $self->pack_block( TYPE_SHB,
      "i s s $fmt_s64",
      PCAPNG_MAGIC,
      1, 0,          # maj,min vers
      pack_s64(-1),  # total length - unknown as yet
   );
}

sub read_header
{
   my $self = shift;

   my ( $type, $magic, $vermajor, $verminor, $len ) =
      $self->unpack_block( "i s s $fmt_s64" );

   die "Bad header" unless $type == TYPE_SHB;

   die "Bad magic - $magic" unless $magic == PCAPNG_MAGIC;
   die "Bad version - $vermajor.$verminor" unless $vermajor == 1 and $verminor == 0;
}

sub write_iface_block
{
   my $self = shift;
   my ( $info ) = @_;

   my $linktype = $info->{analyser}->PCAP_DLT;
   defined $linktype or
      die "TODO: Not sure what pcap linktype for $info->{analyser}\n";

   $self->pack_block( TYPE_IDB,
      "s x[s] i a*",
      $linktype, 65536,

      $self->_pack_options( 
         IF_NAME()    => $info->{ifname},
         IF_TSRESOL() => pack( "C", 6 ),
         # TODO: consider other options; IPv4/6 addr, MAC addr, ...
      ),
   );

   return $self->{next_interface_id}++ || 0;
}

sub handle_iface_block
{
   my $self = shift;
   my ( $content ) = @_;

   my ( $linktype, $maxlen, $opts ) = unpack( "s x[s] i a*", $content );

   my $analyser = PktDump::Analyser->for_pcap_dlt( $linktype );

   my %opts = $self->_unpack_options( $opts );

   my $interface_id = $self->{next_interface_id}++ || 0;

   $self->{interface_id2info}{$interface_id} = {
      # These IDs are NOT kernel's ones but that doesn't matter
      ifindex  => $interface_id,
      ifname   => $opts{IF_NAME()} || "?",
      analyser => $analyser,
      hatype   => $analyser->HATYPE,
   };
}

my %dir2flag = (
   RX => 1, # inbound
   TX => 2, # outbound
);
my %flag2dir = reverse %dir2flag;

my %scope2flag = (
   U => 1, # unicast
   M => 2, # multicast
   B => 3, # broadcast
   O => 4, # otherhost
);
my %flag2scope = reverse %scope2flag;

sub write_packet
{
   my $self = shift;
   my ( $info, $packet ) = @_;

   # pcapng file numbers interfaces implictly by ordering of the interface
   # definition blocks

   my $interface_id;

   unless( defined( $interface_id = $self->{ifindex2id}{ $info->{ifindex} } ) ) {
      $interface_id = $self->write_iface_block( $info );

      $self->{ifindex2id}{ $info->{ifindex} } = $interface_id;
   }

   my ( $usec_lo, $usec_hi );
   {
      use bigint;
      my $usec = $info->{ts_sec} * 1_000_000 + $info->{ts_usec}; # units of 10^-6 seconds
      $usec_lo = $usec & 0xffffffff;
      $usec_hi = $usec >> 32;
   }

   my $flags =
      ( $dir2flag{$info->{dir}}           || 0 ) |
      ( $scope2flag{$info->{scope} || ""} || 0 ) << 2;

   $self->pack_block( TYPE_EPB,
      "I I I I I a* x![I] a*",
      $interface_id,
      $usec_hi, $usec_lo,
      length $packet, length $packet,
      $packet,

      $self->_pack_options(
         EPB_FLAGS() => pack( "i", $flags ),
      ),
   );
}

sub read_packet
{
   my $self = shift;

   while(1) {
      my ( $type, $content ) = $self->read_block
         or return undef;

      if( $type == TYPE_EPB ) {
         my ( $interface_id, $usec_hi, $usec_lo, $caplen, $datalen ) = 
            unpack( "I I I I I", $content );
         substr( $content, 0, 20 ) = "";

         ( $_[0], my $opts ) = unpack( "a$datalen x![I] a*", $content );

         my %opts = $self->_unpack_options( $opts );

         my $info = { %{ $self->{interface_id2info}{$interface_id} } };

         my $flags = unpack( "i", $opts{EPB_FLAGS()} );
         $info->{dir}   = $flag2dir  { $flags      & 0x00000003 };
         $info->{scope} = $flag2scope{ $flags >> 2 & 0x00000007 };

         {
            use bigint;
            my $usec = $usec_hi << 32 | $usec_lo;

            $info->{ts_sec}  = int( $usec / 1000000 );
            $info->{ts_usec} = $usec % 1_000_000;
         }

         # TODO: ethertype, addr

         return $info;
      }
      elsif( $type == TYPE_IDB ) {
         $self->handle_iface_block( $content );
      }
      else {
         die "TODO: Not sure how to handle block type $type\n";
      }
   }
}

1;
