package PktDump::PacketSocket;

use strict;
use warnings;

use base qw( PktDump::PacketSourceRole );

use IO::Socket::Packet 0.07;
use Socket::Packet qw(
   pack_sockaddr_ll
   ETH_P_ALL
   /^PACKET_/
);

use Linux::SocketFilter qw( attach_filter );

use PktDump::Analyser;

sub new
{
   my $class = shift;
   my %args = @_;

   my $sock = IO::Socket::Packet->new(
      Protocol => 0, # don't capture packets yet until we've set up the filter
   ) or die "Cannot create PF_PACKET socket - $!";

   my $self = bless {
      sock => $sock,
   }, $class;

   if( defined $args{filter} ) {
      my $analyser = PktDump::Analyser->for_hatype( $sock->hatype || 1 );
      my $filter = $analyser->gen_filter( $args{filter} );

      attach_filter( $sock, $filter ) or
         die "Cannot SO_ATTACH_FILTER - $!";
   }

   # 2048 bytes is sufficient for an Ethernet frame.
   # 128 frames should be ample buffer space
   # mmap() in blocks of 64KiB
   if( eval { $sock->setup_rx_ring( 2048, 128, 2**16 ) } ) {
      bless $self, __PACKAGE__."::RXRing";
   }

   my $ifindex = 0;
   if( $args{iface} ) {
      $ifindex = $sock->ifname2index( $args{iface} );
   }

   $sock->bind( pack_sockaddr_ll( ETH_P_ALL, $ifindex, 0, 0, '' ) ) or
      die "Cannot bind() - $!";

   return $self;
}

sub finalise
{
   my $self = shift;
   my $sock = $self->{sock};

   my $stats = $sock->statistics;

   printf "Captured %d packets, dropped %d\n", $stats->{packets}, $stats->{drops};
}

my %ifname_cache;
sub ifindex2name
{
   my $self = shift;
   my ( $ifindex ) = @_;
   return $ifname_cache{$ifindex} ||= $self->{sock}->ifindex2name( $ifindex );
}

sub _pkttype2info
{
   my $self = shift;
   my ( $pkttype ) = @_;

   return 
      dir =>   ( $pkttype == PACKET_OUTGOING  ) ? "TX" : "RX",
      scope => ( $pkttype == PACKET_HOST      ) ? "U" :
               ( $pkttype == PACKET_MULTICAST ) ? "M" :
               ( $pkttype == PACKET_BROADCAST ) ? "B" :
               ( $pkttype == PACKET_OTHERHOST ) ? "O" : undef;
}

sub read_packet
{
   my $self = shift;
   my ( undef ) = @_; # buffer

   my $sock = $self->{sock};

   my ( $proto, $ifindex, $hatype, $pkttype, $addr ) = $sock->recv_unpack( $_[0], 8192, 0 )
      or return undef;
   my ( $ts_sec, $ts_usec ) = $sock->timestamp;

   my $ifname = $ifname_cache{$ifindex} ||= $sock->ifindex2name( $ifindex );

   my $info = {
      ethertype => $proto,
      ifindex   => $ifindex,
      ifname    => $self->ifindex2name( $ifindex ),
      hatype    => $hatype,
      pkttype   => $pkttype,
      addr      => $addr,
      ts_sec    => $ts_sec,
      ts_usec   => $ts_usec,

      analyser => PktDump::Analyser->for_hatype( $hatype ),

      $self->_pkttype2info( $pkttype ),
   };

   return $info;
}

package PktDump::PacketSocket::RXRing;

use base qw( PktDump::PacketSocket );

sub recv_packets
{
   my $self = shift;
   my ( $handle_pkt ) = @_;

   my $sock = $self->{sock};

   my $sigint = $SIG{INT};
   my $intr;
   local $SIG{INT} = sub { $sigint->(); $intr = 1; die };

   eval { 
      while( defined $sock->wait_ring_frame( my $packet, \my %info ) ) {
         my $pkttype = $info{sll_pkttype};

         if( $info{sll_hatype} == 1 and $info{tp_vlan_tci} ) {
            # Kernel has mangled the VLAN header in an Ethernet packet.
            # Header now:     [12 byte eth. addrs] [ethertype] [L3...]
            # Header we want: [12 byte eth. addrs] [802.1Q ethertype] [VLAN TCI] [ethertype] [L3...]
            # Copy the buffer since currently this is kernel mmap()ed
            $packet = substr( $packet, 0, 12 ) . # Ethernet addrs
                      pack( "n n", 0x8100, $info{tp_vlan_tci} ) .
                      substr( $packet, 12 );     # Ethertype, L3
            $info{ethertype} = 0x8100;
         }

         $handle_pkt->( $packet, {
            ethertype => $info{sll_protocol},
            ifindex   => $info{sll_ifindex},
            ifname    => $self->ifindex2name( $info{sll_ifindex} ),
            hatype    => $info{sll_hatype},
            pkttype   => $pkttype,
            addr      => $info{sll_addr},
            ts_sec    => $info{tp_sec},
            ts_usec   => int( $info{tp_nsec} / 1000 ),

            analyser => PktDump::Analyser->for_hatype( $info{sll_hatype} ),

            $self->_pkttype2info( $pkttype ),
         } ) or last;

         $sock->done_ring_frame;
      }
   };
   die $@ unless $intr;
}

1;
